#
# $Id: inetd.tcl,v 1.9 2000/03/21 17:08:08 elias Exp $
# Copyright 1999 Security Dynamics
#
# Utils to modify services /etc/inetd.conf.  Wraps them using the
# "iwrap" program.
#
# Usage:
#	IWrap::Wrap <service> <path/to/iwrap> <result>
#	IWrap::Unwrap <service> <path/to/iwrap> <result>
#
# Both return 1 if successful, 0 if not.  On error, the result
# variable will be a list containing a symbolic error and an error
# message.  Sometimes there is an extra element.
#
# Errors are:
#	NOTFOUND	 service not found
#	NOTWRAPPED	 service is not wrapped
#	ISWRAPPED	 service is already wrapped
#	BADARGS		 canot unwrap - line has been modified
#	CANTWRITE + file cannot write file
#	CANTREAD + file	 cannot read file
#	NOCHAR		 automatic wrapping failed (very unusual)
#

namespace eval ::IWrap { ;#

namespace export Wrap Unwrap Find CanAccess

#----------------------------------------------------------------------
# Private procs

# Follow symlinks for /etc/inetd.conf
#
proc findfile {file} {
    while {[file type $file] == "link"} {
	set link [file readlink $file]
	if ![string match /* $link] {
	    set link [file join [file dirname $file] $link]
	}
	set file $link
    }
    return $file
}

# Find the line for a service and return the whole file
#
proc findservice {service resultlist {file /etc/inetd.conf}} {
    upvar $resultlist filedata

    set file [findfile $file]
    if [catch {set f [open $file]} msg] {
	return [list CANTREAD $file $msg]
    }
    set filedata [split [read $f] "\n"]
    close $f

    set n -1
    foreach line $filedata {
	incr n
	if {[string length $line] == 0} continue
	if [string match #* $line] continue

	foreach {svc stream tcp nowait user} $line break
	if [string compare $svc $service] continue
	if [string compare $tcp tcp] {
            if [string compare $tcp tcp6] continue
        }
	if [string compare $stream stream] {
	    set filedata "$service is not of type \"stream\""
	    return -1
	}
	if [string compare $nowait nowait] {
	    set filedata "$service does not have \"nowait\" status"
	    return -1
	}
	set resultlist $filedata
	return $n
    }
    set filedata "cannot find service $service in $file"
    return -1
}

# Check if a line is already wrapped
#
proc iswrapped {line iwrap} {
    set prog [lindex $line 5]
    return [string match */[file tail $iwrap] $prog]
}



proc listfile_add {listname element file {umsg ""}} {
    if ![file exists $file] {
        if [catch {set f [open $file w]} msg] {
            set result [list CANTWRITE $msg $file]
            return 0
        }
	puts $f $umsg
        puts $f "set $listname {$element}"
        close $f
    } else {
        source $file
        if ![info exists $listname] {
            set $listname {}
        }
        if {[lsearch -exact $[set $listname] $element] == -1} {
            lappend $listname $element
            if [catch {set f [open $file w]} msg] {
                set result [list CANTWRITE $msg $tmp]
                return 0
            }
	    puts $f $umsg
            puts $f "set $listname {[set $listname]}"
            close $f
        }
    }
    return 1
}


proc listfile_del {listname idx file {umsg ""}} {
    if [file exists $file] {
        source $file
        set f [open $file w]
        set $listname [lreplace [set $listname] $idx $idx] 
	if {[llength [set $listname]] > 0} {
	    if [catch {set f [open $file w]} msg] {
		return 0;
	    }
	    puts $f $umsg
	    puts $f "set $listname {[set $listname]}"
	    close $f
	} else {
	    catch {file delete -force $file}
	}
        
        return 1;
    } else {
        return 0;
    }
}

# Wrap a line
#
proc wrapline {line resultvar iwrap {wrapargs {}}} {
    upvar $resultvar result

    if [iswrapped $line $iwrap] {
	set result [list ISWRAPPED "already wrapped"]
	return 0
    }
    set base [file tail $iwrap]
    foreach {svc stream tcp nowait user} $line break
    set prog [lindex $line 5]
    set args [lrange $line 6 end]

    set arglist [concat $prog $args]
    foreach c {: , + % - ~} {
	if {[string first $c $arglist] < 0} {
	    set sep $c
	    break
	}
    }
    if ![info exists sep] {
	set result [list NOCHAR "cannot find a harmless separator character"]
	return 0
    }
    set umsg " #\n\
	    # Auto generated.\n\
	    # Wrapped sevices to be restored to tcp6 in inetd.conf\n\
	    #\n"
    if {![string compare "tcp6" $tcp]} {
	if {![listfile_add tcp6_entries $svc [CatDir VAR]/tcp6_entries $umsg]} {
	    set result [list CANTWRITE "cannot record tcp flavor for entry"]
	    return 0
	}
	set line [lreplace $line 2 2 "tcp"]
    }
    set sockarg [CatDir VAR]/.catsock/s%
    set arglist "$wrapargs -s$sockarg -ac$sep[join $arglist $sep]"
    set result [concat [lrange $line 0 4] [list $iwrap $base $arglist]]
    set result [join $result "\t"]

    return 1
}

# Unwrap a line that must have been wrapped by wrapline
#
proc unwrapline {line resultvar iwrap} {
    upvar $resultvar result

    set base [file tail $iwrap]
    set prog [lindex $line 5]

    if ![string match */$base $prog] {
	set result [list NOTWRAPPED "not wrapped"]
	return 0
    }

    set umsg " #\n\
	    # Auto generated.\n\
	    # Wrapped sevices to be restored to tcp6 in inetd.conf\n\
	    #\n"
    if [file exists [CatDir VAR]/tcp6_entries] { 
        source [CatDir VAR]/tcp6_entries
    } else {
	set tcp6_entries {}
    }
    set service [lindex $line 0]
    set idx [lsearch -exact $tcp6_entries $service]
    if {$idx >= 0} {
        set line [lreplace $line 2 2 tcp6]
	listfile_del tcp6_entries $idx [CatDir VAR]/tcp6_entries $umsg
    }
    set result [lrange $line 0 4]
    set s [lindex $line end]
    if ![string match -ac* $s] {
	set result [list BADARGS "wrap args modified: unable to unwrap"]
	return 0
    }
    set c [string index $s 3]
    set s [string range $s 4 end]
    set result [join [concat $result [split $s $c]] "\t"]

    return 1
}

# Tell inetd to re-read its config file
#
proc notify {} {
    exec sh -c { refresh -s inetd }
}

# help proc for Wrap and Unwrap
#
proc modify {func service resultvar iwrapprog file} {
    upvar $resultvar result

    set idx [findservice $service data $file]
    if {$idx < 0} {
	set result [list NOTFOUND "service $service not found in $file"]
	return 0
    }
    set line [lindex $data $idx]
    if ![$func $line result $iwrapprog] {
	return 0
    }
    set data [lreplace $data $idx $idx $result]
    set file [findfile $file]
    set tmp $file.[pid]
    catch {file delete -force -- $tmp}
    if [catch {set f [open $tmp w]} msg] {
	set result [list CANTWRITE $msg $tmp]
	return 0
    }
    puts $f [string trim [join $data "\n"]]
    close $f
    if ![file exists $file.bak] {
	catch {file copy -force -- $file $file.bak}
    }
    set attrs [file attributes $file]
    if [catch {file rename -force -- $tmp $file} msg] {
	set result [list CANTWRITE $msg $file]
	catch {file delete -force -- $tmp}	
	return 0
    }
    catch {eval file attributes $file $attrs}
    catch { notify }

    return 1
}

#----------------------------------------------------------------------
# Public procs

# Wrap service wrapprog resultvar [filename]
# -- iwrap a service in inetd.conf
#
proc Wrap {service iwrapprog resultvar {file /etc/inetd.conf}} {
    upvar $resultvar result

    return [modify wrapline $service result $iwrapprog $file]
}

# Unwrap service wrapprog resultvar [filename]
# -- undo what Wrap did
#
proc Unwrap {service iwrapprog resultvar {file /etc/inetd.conf}} {
    upvar $resultvar result

    return [modify unwrapline $service result $iwrapprog $file]
}

# Find
# -- Locate service in inetd.conf, and return as a list, with
#    everything from the program and onwards as a single element
#
proc Find {service {file /etc/inetd.conf}} {
    set idx [findservice $service data $file]
    if {$idx < 0} {
	return {}
    }
    set line [lindex $data $idx]
    
    set result [lrange $line 0 4]
    lappend result [lrange $line 5 end]
}

# CanAccess
# -- See if the inetd.conf file looks writable
#
proc CanAccess {{file /etc/inetd.conf}} {
    set file [findfile $file]

    if [catch {set f [open $file r+]}] {
	return 0
    }
    close $f
    return 1
}

}; # end namespace IWrap


