#
# $Id: confcmd.tcl,v 1.22 2000/05/05 15:35:04 elias Exp $
#



catch {namespace import CatInfo::*}
catch {namespace import Menu::*}


# 
# argspec is a specification of the parameters (with variables
# to be filled in) and an argv.
#
# argspec is a list of tuples on the form;
#   cmd     - the argument string to search for
#   type    - Single character type decl of the argument value.
#             Omitted type character indicates no value of argument
#             n - none
#             i - integer
#             s - string
#             h - hostname on the form <address/host>[:<port>]
#   val     - The value of the parameter. If no type is given, val
#             will be set to '1' if the argument is present in argv
#   present - Set to 1 if the parameter was present in argv, 0 otherwise.
#             Sort of redundant but simplifies code.
#           
# The present variable is guaranteed to be set, val is not.
# If parsing of the arguments fails, 0 is returned. 1 otherwise.

proc getargs {argspec argv {custom {}}} {
    global Menus
    set do_custom 0
    if [info exists $custom] { 
	upvar $custom cparams
	set do_custom 1
    }
    foreach {cmd type val present} $argspec {
	upvar $val v
	upvar $present p
	set p 0
	set v ""
	if {$type != "n"} {
	    set idx [lsearch -regexp $argv "$cmd=*"]
	    if {$idx >= 0} {
		set p 1
		regexp {[a-zA-Z0-9_]*=(.+)} [lindex $argv $idx] match v
		#
		# Check type of arg
		switch -exact -- $type {
		    "i" { 
			if ![regexp {^[0-9]+$} $v tmp] {
			    return 0
			}
		    }
		    "s" {
			if {[string length $v] <= 0} {
			    return 0
			}
		    }
		    "a" {
			if ![LooksLikeHost] {
			    return 0
			}
		    }
		    "p" {
			set protolist $Menus(protolist)
			if {[lsearch -exact $protolist $v] < 0} {
			    return 0
			}
		    }
		}
		set argv [lreplace $argv $idx $idx]
	    }
	} else {
	    set idx [lsearch -exact $argv "$cmd"]
	    if {$idx >= 0} {
		set p 1
		set v 1
		set argv [lreplace $argv $idx $idx]
	    }
	}
    }

    #
    # Did we recognize everything in argv?
    # If not and we have custom list, add all remaining var=values
    # to the custom list.

    if {[llength $argv] > 0} {
	if $do_custom {
	    set var ""
	    set val ""
	    foreach varval $argv {
		regexp {([^=]*)=(.*)} $varval match var val
		if {[string compare $var ""] && [string compare $val ""]} {
		    lappend cparams $var $val
		} else {
		    return 0
		}
	    }
	    return 1
	} else {
	    return 0
	}
    } else {
	return 1
    }
}

proc AddCmdUsage {} {
    global Menus
    return "Usage: $Menus(add,usage)"
}
	
proc AddCmd {argv} {
    global Menus
    
    set argspec {
	"help"   n help    help_p
	"svc"    s service service_p
	"inetd"  s itag    itag_p
	"port"   i port    port_p
	"server" h server  server_p
	"method" s method  method_p
	"desc"   s desc    desc_p
	"proto"  p proto   proto_p
    }
    set custom {}
    if {![getargs $argspec $argv custom]} {
	return -1
    }
    if $help_p {
	puts "[AddCmdUsage]"
	puts "$Menus(add,help)"
	return 0
    }
    if {!$service_p || !$method_p || !$proto_p} {
	return -1
    }

	
    if {$itag_p && !$port_p && !$server_p} {
  	set args [list $itag $service $proto $method auto $desc]
	if ![eval ConfigureInetdPort $args] {
	    puts stderr "Failed to configure port"
	    return -2
	}
	set port [GetServByName $itag]
    } elseif {!$itag_p && $port_p && $server_p } {
	GetConfiguredPorts data
	if [info exists data($port,service)] {
	    puts stderr  "Port already configured."
	    return -2
	}
	set args [list $port $service $proto $method $server auto $desc]
	if ![eval ConfigurePort $args] {
	    puts stderr "Failed to configure port"
	    return -2
	}
    } else {
	return -1
    }
    set lib [CatDir APP]/lib
    if [file exists $lib/$service/config.tcl] {
	puts " $lib/$service/config.tcl did exist"
	catch { namespace delete ::CatModule }
	catch { source $lib/$service/config.tcl } 
	
	if [catch { CatModule::ModuleSpecificConf data }] {
	    puts "Error configuring module specific data"
	    return -2
	}
    }
    #
    # Set custom parameters
    foreach {path val} $custom {
	Set .tcp.s$port.$path $val
    }
    ConfFlush
    return 0
}

proc DelCmdUsage {} {
    global Menus
    return "Usage: $Menus(del,usage)"
}

proc DelCmd {argv} {
    global Menus
    set argspec {
	"help" n help help_p
	"port" i port port_p
    }
    if {![getargs $argspec $argv]} {
	return -1
    }
    if $help_p {
	puts "[DelCmdUsage]"
	puts "$Menus(del,help)"
	return 0
    }
    if !$port_p {
	return -1
    }
    GetConfiguredPorts data
    if ![info exists data($port,service)] {
	puts stderr  "Port not configured."
	return -2
    }
    GetPortInfo p $port
    if [catch {DeletePort $port} ret] {
	puts stderr "Could not delete port."
	puts stderr "\t($ret)"
	return -2
    } else {
	puts "Deleted \"$p(service)\" on port $port"
    }

    return 0
}

proc ListCmdUsage {} {
    global Menus
    return "Usage: $Menus(list,usage)"
}

proc ListCmd {argv} {
    global Menus
    set argspec {"help" n help help_p}
    if {![getargs $argspec $argv]} {
	return -1
    }
    if $help_p {
	puts "[ListCmdUsage]"
	puts "$Menus(list,help)"
	return 0
    }
    if ![ShowConfiguredPorts] {
	puts "\nNo ports configured"
    }
    return 0
}

proc HostPsdCmdUsage {} {
    global Menus
    return "Usage: $Menus(hostpsd,usage)"
}

proc HostPsdCmd {argv} {
    global Menus
    set argspec {
	"help"    n help    help_p 
	"vcfile"  s vcfile  vcfile_p
	"pwdfile" s pwdfile pwdfile_p
    }
    if {![getargs $argspec $argv]} {
	return -1
    }    
    if $help_p {
	puts "[HostPsdCmdUsage]"
	puts "$Menus(hostpsd,help)"
	return 0
    }
    if {!$vcfile_p || !$pwdfile_p} {
	return -1
    }
    if ![ConfHostPSD $vcfile $pwdfile] {
	puts stderr "Could not set PSD."
	return -2
    }
    ConfFlush
    return 0
}

proc CaCmdUsage {} {
    global Menus
    return "Usage: $Menus(ca,usage)"
}

proc CaCmd {argv} {
    global Menus
    set argspec {
	"help"   n help   help_p
	"list"   n list   list_p
	"del"    i delidx del_p
	"import" s cafile import_p
    }
    if {![getargs $argspec $argv]} {
	return -1
    }    
    if $help_p {
	puts "[CaCmdUsage]"
	puts "$Menus(ca,help)"
	return 0
    }
    Get .global.ca.dbfile dbfile
    if { $list_p && !$del_p && !$import_p} {
	ShowImportedCA 
    } elseif {!$list_p && $del_p && !$import_p} {
	catch {set l [exec cadb -d $dbfile -l 2>@ stderr]}
	set l [split $l "\n"]
	catch {set dn [lindex $l $delidx]}
	catch {exec cadb -v -d $dbfile -r $dn 2>@ stderr}
    } elseif {!$list_p && !$del_p && $import_p} {
	catch {exec cadb -v -d $dbfile -a $cafile 2>@ stderr}
    } else {
	return -1
    }

    return 0
}

proc BoksCmdUsage {} {
    global Menus
    return "Usage: $Menus(boks,usage)"
}

proc BoksCmd {argv} {
    global Menus
    set defdomain 6500
    set argspec {
	"help"    n help    help_p
	"nodekey" s nodekey nodekey_p
	"domain"  i domain  domain_p 
	"servers" s servers servers_p
    }
    if {![getargs $argspec $argv]} {
	return -1
    }    
    if $help_p {
	puts "[BoksCmdUsage]"
	puts "$Menus(boks,help)"
	return 0
    }
    
    GetBoksInfo info
    if !$domain_p {
	set domain $defdomain
    }
    ConfBoks domain $domain
    
    if {$nodekey_p} {
	if ![PasswordFile boks where] {
	    puts stderr "Cannot write BoKS Node Key File"
	    return -2
	}
	set f [open $where w]
	puts $f $nodekey
	close $f
	ConfBoks nodekeyfile $where
    }
	
    if {$servers_p} {
	if {[string compare $servers "."] == 0} {
	    ConfBoks servers ""
	} else {
	    # replace all ; to spaces
	    
	    regsub -all {;} $servers " " servlist
	    ConfBoks servers $servlist
	}
    }
    ConfFlush
    return 0
}

proc ElsCmdUsage {} {
    global Menus
    return "Usage: $Menus(els,usage)"
}

proc ElsCmd {argv} {
    global Menus
    set argspec {
	"help"    n help    help_p
	"servers" s servers servers_p
    } 
    if {![getargs $argspec $argv]} {
	return -1
    }
    if $help_p {
	puts "[ElsCmdUsage]"
	puts "$Menus(els,help)"
	return 0
    }
    GetELSInfo els
    if {$servers_p} {
	if {[string compare $servers "."] == 0} {
	    ConfELS ""
	} else {
	    # replace all ; to spaces
	    regsub -all {;} $servers " " servlist
	    ConfELS $servlist
	}
    }
    ConfFlush
    
    return 0
}

proc ShowCmdUsage {} {
    global Menus
    return "Usage: $Menus(show,usage)"
}
    
proc ShowCmd {argv} {
    global Menus
    set argspec {"help" n help help_p}
    if {![getargs $argspec $argv]} {
	return -1
    }
    if $help_p {
	puts "[ShowCmdUsage]"
	puts "$Menus(show,help)"
	return 0
    }
    ShowCurrentConfig
    return 0
}

proc SaveCmdUsage {} {
    global Menus
    return "Usage: $Menus(save,usage)"
}

#
# Generate a TCL-script that can be sourced and that inserts
# saved config data. If no "autofile" arg is given, output goes
# to stdout.
# savespec is a list that may contain the strings "vc", "ca" and "cfg"
#   "vc" - save the virtual card
#   "ca" - save the CA database
#   "cfg" - save data in configuration files
#
# Returns 0 on sucess or 1 on error

proc DumpConfig {version rootdir savespec {autofile ""}} {
    set save_vc 0
    set save_ca 0
    set save_cfg 0

    # see what we should save
    foreach {spec} $savespec {
	set save_$spec 1
    }

    CatInfo::Init $rootdir
    
    set globlist [CatInfo::List .global]
    set servlist [CatInfo::List .services]
    set portlist [CatInfo::List .tcp]

    set list [concat $globlist $servlist $portlist]

    #
    # get the host PSD/password and save them in vars
    if {$save_vc} {
	set hostpsd ""
	if {[CatInfo::Get .global.hostpsd.file psdfile] == 0} {
	    puts stderr "No Agent Host Virtual Card set. Skipping."
	} else {
	    if [catch {open $psdfile RDONLY} fileId] {
		puts stderr "Skipping Host Virtual Card $fileId"
		puts stderr "\t($fileId)"
	    } else {
		while {[gets $fileId line] >= 0} {
		    set hostpsd "$hostpsd$line"
		}
		close $fileId
	    }
	}
	#
	# Save the PSD password if stored in file.
	# (will be saves in configuration values otherwise)
	set psdpsw ""
	if {[CatInfo::Get .global.hostpsd.passwordfile psdpswfile] == 0} {
	    puts stderr "No Agent Host Virtual Card password file set. Skipping."
	} else {
	    if [catch {open $psdpswfile RDONLY} fileId] {
		puts stderr "Skipping Host Virtual Card Password"
		puts stderr "\t($fileId)"
	    } else {
		while {[gets $fileId line] >= 0} {
		    set psdpsw "$psdpsw$line"
		}
		close $fileId
	    }
	}
    }
    
    #
    # get the CA db and put it in a var
    if {$save_ca} {
	set cadb ""
	if {[CatInfo::Get .global.ca.dbfile cadbfile] == 0} {
	    puts stderr "No CA database set. Skipping."
	} else {
	    if [catch {open $cadbfile RDONLY} fileId] {
		puts stderr "Skipping CA database file"
		puts stderr "\t($fileId)"
	    } else {
		while {[gets $fileId line] >= 0} {
		    set line "$line\n"
		    set cadb "$cadb$line"
		}
		close $fileId
	    }
	}
    }
    
    #
    # open and initialize the backup file
    if {$autofile != ""} {
	if [catch {open $autofile {WRONLY CREAT TRUNC}} fileId] {
	    puts stderr "Cannot open $autofile: $fileId"
	    return -2
	}
    } else {
	set fileId stdout
    }

    #
    # Write version
    puts $fileId "set autogenVersion \{$version\}"

    #
    # Write the hostpsd and password
    if {$save_vc} {
	puts $fileId "set autogenHostpsd \{$hostpsd\}"
	puts $fileId "set autogenPsdpwd \{$psdpsw\}"
    } else {
	puts $fileId "set autogenHostpsd \"\""
	puts $fileId "set autogenPsdpwd \"\""
    }

    #
    # write the cadb
    if {$save_ca} {
	puts $fileId "set autogenCaDb \{$cadb\}"
    } else {
	puts $fileId "set autogenCaDb \"\""
    }

    #
    # Write the list of currently installed services
    puts $fileId "set autogenInstServ {"
    if {$save_cfg} {
	set services [GetInstalledServices sdata]
	puts $fileId $services
    }
    puts $fileId "}"

    #
    # Write list of all config values
    puts $fileId "set autogenCfgList {"
    if {$save_cfg} {
	foreach elem $list {
	    CatInfo::Get $elem val
	    puts $fileId "$elem {$val}"
	}
    }
    puts $fileId "}"
    
    if {$autofile != ""} {
	close $fileId
    }
    return 0
}

proc SaveCmd {argv} {
    global Menus
    set version "1.0"
    set savespec {"vc" "ca" "cfg"}
    set argspec {
	"help" n help help_p 
	"vc"   n vc   vc_p
	"ca"   n ca   ca_p
	"cfg"  n cfg  cfg_p
	"file" s file file_p
    }
    if {![getargs $argspec $argv]} {
	return -1
    }
    if $help_p {
	puts "[SaveCmdUsage]"
	puts "$Menus(save,help)"
	return 0
    }
    if { $vc_p || $ca_p || $cfg_p } {
	set savespec {}
	if {$vc_p} {
	    lappend savespec "vc"
	}
	if {$ca_p} {
	    lappend savespec "ca"
	}
	if {$cfg_p} {
	    lappend savespec "cfg"
	}
    }

    if !$file_p {
	set file ""
    }
    set ret [DumpConfig $version [CatDir ETC] $savespec $file]
    return $ret
}

proc RestCmdUsage {} {
    global Menus
    return "Usage: $Menus(restore,usage)"
}

proc RestCmd {argv} {
    global Menus
    set argspec {
	"help" n help help_p 
	"file" s file file_p
    }
    if {![getargs $argspec $argv]} {
	return -1
    }
    if $help_p {
	puts "[RestCmdUsage]"
	puts "$Menus(restore,help)"
	return 0
    }
    if !$file_p {
	return -1
    }
    if [catch {source $file} result] {
	puts stderr "$result"
	return -2
    }

    CatInfo::Init [CatDir ETC]

    if {[llength $autogenCfgList] > 0} {
	#
	# Remove all current port configurations
	set currports [GetConfiguredPorts data]
	foreach port $currports {
	    DeletePort $port
	}

	#
	# Remove all config params for the services that 
	# was present when the configuration was saved
	# We need to keep the values added after the configuration
	# was saved. Otherwise, no Agents of those services can be 
	# configured
	foreach {oldserv} $autogenInstServ {
	    set oldservlist [CatInfo::List .services.$oldserv]
	    foreach {oldservval} $oldservlist {
		CatInfo::Set "$oldservval" "" 1
	    }
	}

	#
	# Remove the global and the port specific config values
	# These are to be replaced with the values in the
	# save file
	set globlist [CatInfo::List .global]
	set portlist [CatInfo::List .tcp]
	set actualvars [concat $globlist $portlist]
	
	array set bkpvars $autogenCfgList
	
	foreach {var} $actualvars {
	    if {[array names bkpvars $var] == ""} {
		CatInfo::Set $var "" 1
	    }
	}

	#
	# Loop to set all backup values in the list
	foreach {var val} $autogenCfgList {
	    CatInfo::Set $var $val
	}

	#
	# Install the ports from the saved configuration
	set portnolist {}
	GetConfiguredPorts portarray
	if [array exists portarray] {
	    set sid [array startsearch portarray]
	    while {[set str [array nextelement portarray $sid]] != ""} {
		regsub {(^[0-9]+),.*} $str {\1} portno
		if {[lsearch -exact $portnolist $portno] < 0} {
		    lappend portnolist $portno
		}
	    }
	}
	foreach {port} $portnolist {
	    CatInfo::Get .tcp.s$port.service service
	    CatInfo::Get .tcp.s$port.protocol protocol
	    CatInfo::Get .tcp.s$port.authmethod method	    
	    CatInfo::Get .tcp.s$port.start start
	    CatInfo::Get .tcp.s$port.description descr
	    CatInfo::Get .tcp.s$port.inetd inetd
	    #
	    # Save the port specific data. Confguring a port
	    # will set all Agent specific conf data to default values
	    # and we must reset that afterwards.
	    set portparamlist {}
	    set portparams [CatInfo::List .tcp.s$port]
	    foreach portparam $portparams {
		CatInfo::Get $portparam portparamval
		lappend portparamlist $portparam $portparamval
	    }
	    # configure ports
	    if [CatInfo::Get .tcp.s$port.inetd inetd] {
		ConfigureInetdPort $inetd $service $protocol $method $start $descr
	    } else {
		CatInfo::Get .tcp.s$port.server server
		regsub {^tcp:} $server "" server
		ConfigurePort $port $service $protocol $method $server $start $descr
	    }
	    #
	    # Reset Agent specific config values
	    foreach {path val} $portparamlist {
		CatInfo::Set $path $val
	    }
	}

	CatInfo::Flush
    }

    #
    # Writeback of hostpsd, psd password and cadb
    if {$autogenHostpsd != ""} {
	if {[CatInfo::Get .global.hostpsd.file psdfile] == 0} {
	    puts stderr "No Agent Host Virtual Card file set. Skipping."
	} else {
	    if [catch {open $psdfile {WRONLY CREAT TRUNC}} fileId] {
		puts stderr "Cannot write $psdfile: $fileId"
		return -2
	    } 
	    puts $fileId $autogenHostpsd
	    close $fileId
	}
    }
    if {$autogenPsdpwd != ""} {
	if {[CatInfo::Get .global.hostpsd.passwordfile psdpwdfile] == 0} {
	    puts stderr "No Agent Host VC password file set. Skipping."
	} else {
	    if [catch {open $psdpwdfile {WRONLY CREAT TRUNC}} fileId] {
		puts stderr "Cannot write $psdpwdfile: $fileId"
		return -2
	    } 
	    puts $fileId $autogenPsdpwd
	    close $fileId
	}
    }
    
    if {$autogenCaDb != ""} {
	if {[CatInfo::Get .global.ca.dbfile cadbfile] == 0} {
	    puts stderr "No CA database file set. Quitting."
	} else {
	    if [catch {open $cadbfile {WRONLY CREAT TRUNC}} fileId] {
		puts stderr "Cannot write $cadbfile: $fileId"
		return -2
	    } 
	    puts $fileId $autogenCaDb
	    close $fileId
	}
    }
    

    return 0
}
