#
# $Id confmenus.tcl,v 1.4 1999/05/30 09:08:41 mikko Exp $
#
#
#
#

namespace eval ::CatConf {; #

namespace export GetConfiguredPorts ShowConfiguredPorts GetPortInfo
namespace export DeletePort ConfigurePort InstallPortTemplate
namespace export ConfigureInetdPort ConfigureBindHookPort
namespace export CatDir Startup ConfFlush
namespace export ConfHostPSD
namespace export GetBoksInfo ConfBoks ConfBoksNodekey TryBoks
namespace export ConfELS GetELSInfo
namespace export ShowCurrentConfig ShowImportedCA
namespace export GetInstalledServices
namespace export LooksLikeHost LooksLikePort GetServByName HostLooksReachable
namespace export PasswordFile

catch {namespace import ::CatDbg::*}
catch {namespace import ::CatInfo::*}

proc line {len char} {
    set s {}
    for {set i 0} {$i < $len} {incr i} {
	append s $char
    }
    return $s
}

# Return one of the installation directory locations
#
proc CatDir {dir} {
    switch -exact -- $dir {
	APP {
	    return [file join {<<INSTALL:APPDIR>>}]
	}
	ETC {
	    return [file join {<<INSTALL:ETCDIR>>}]
	}
	VAR {
	    return [file join {<<INSTALL:VARDIR>>}]
	}
    }
}

proc PasswordFile {name pathvar {new 0}} {
    upvar $pathvar path

    set path [file nativename [CatDir ETC]/psw.$name]
    if $new {
	if [catch {file delete -force $path} msg] {
	    D 1 "PasswordFile: del $path failed: $msg"
	    return 0
	}
    }
    if [catch {set f [open $path a+ 0600]} msg] {
	D 1 "PasswordFile: write $path failed: $msg"
	return 0
    }
    close $f
    return 1
}

# Start daemon if it is not running already
#
proc Startup {} {
    
    set pidfile [CatDir VAR]/catd.pid

    if {![file exists $pidfile] || [catch {exec start check}]} {
	if [catch {exec start start} msg] {
	    D 1 "failed to start daemon: $msg"
	    return 0
	}
    }
     
    
    return 1
}

# Flush info and notify daemon, if it is running
#
proc ConfFlush {} {
    if ![CatInfo::Flush] {
	return 1
    }

    
    set pidfile [CatDir VAR]/catd.pid

    if ![file exists $pidfile] {
	if ![Startup] {
	    return 0
	}
    } else {
	if [catch {set f [open $pidfile]}] {
	    return 0
	}
	gets $f pid
	close $f
	catch {exec kill -1 $pid}
    }
     

    

    return 1
}


#----------------------------------------------------------------------
# Inet-utils
#

# Check if a string looks like <address/host>[:<port>]
# XXX accepts "_" in names.
#
proc LooksLikeHost {string {port 0} {lecture {}}} {
    set exp {^[-A-Za-z0-9._]+}
    if $port {
	append exp {:\d+}
    }
    append exp {$}
    if [regexp $exp $string] {
	return 1
    }
    if {$port && [regexp {^[-A-Za-z0-9._]+:(.*)} $string {} svc]} {
	if [GetServByName $svc] {
	    return 1
	}
    }
    return 0
}

# Ditto for port numbers
#
proc LooksLikePort {string {min 0}} {
    if {[scan $string %d n] < 0} {
	set n [GetServByName $string]
    }
    if {$n <= $min || $n > 0xffff} {
	return 0
    }
    return $n
}

# Poor man's getservbyname().  It is a bad idea to rely on services
# from a remote location anyway.
#
proc GetServByName {service {proto tcp} {file /etc/services}} {
    if [catch {set f [open $file]} msg] {
	D 1 "faied to open $file: $msg"
	return 0
    }
    set port 0
    set exp {^([a-zA-Z]\S+)\s+(\d+)/}
    append exp $proto
    append exp {\s*([^#]*)}
    while {[gets $f line] > -1} {
	if [regexp $exp $line {} svc p names] {
	    lappend names $svc
	    if {[lsearch -exact $names $service] >= 0} {
		set port $p
		break;
	    }
	}
    }
    close $f

    return $port
}

# Check if it looks possible to connect to a server[:port]
#
proc HostLooksReachable {destination {timeout 5}} {
    global errorCode

    set port 0
    foreach {host port} [split $destination :] break
    set port [LooksLikePort $port]
    if !$port {
	set port 9 ; # discard
    }

    
    
    
    
    
    

    set script {
	puts [pid]
	if [catch {set sock [socket $host $port]}] {
	    foreach {posix code str} $errorCode break
	    puts $code
	}
	puts OK
	close $sock
	exit 0
    }
    if [catch {set p [open "| [info nameofexecutable]" r+]}] {
	return 1
    }
    puts $p "[list set host $host]; [list set port $port]; $script"
    flush $p
    gets $p pid
    fconfigure $p -blocking 0
    set timeout [expr $timeout * 2]
    set s {}
    for {set i 0} {$i < $timeout && [gets $p s] < 0} {incr i} { after 500 }
    set ok [regexp ECONNRESET|ECONNREFUSED|OK $s]
    catch { exec kill $pid }
    catch { close $p }

    
    

    return $ok
}

#----------------------------------------------------------------------
# Port information
#

# Get information on all ports for a service (default: *)
# Service can be a wildcard.  If no version is given, all versions
# are listed. 
#
# Returns the list of ports
#
# The array variable will contain, for each port, $port,$param
# for the parameters: authmethod, start, protocol, server, service
# "serv" (server without the leading tcp:), sdescription (for the
# service) and description (description for the port), inetd (if
# an inetd service)
#
proc GetConfiguredPorts {arrayvar {service {}}} {
    upvar $arrayvar data

    if ![string length $service] {
	set service *
    } elseif ![regexp {.*-\d+\.\d+$} $service] {
	set service $service-*
    }
    set values {authmethod start protocol service server serv inetd}
    set ports(0) 0
    foreach node [List .tcp] {
	if ![regexp {.tcp.s([^.]+).(.*)} $node {} port sub] continue
	if {[lsearch -exact $values $sub] > -1} {
	    Get $node value
	    set data($port,$sub) $value
	}
	if {[string compare $sub service] == 0} {
	    Get $node ports($port)
	} elseif {[string compare $sub server] == 0} {
	    set s {}
	    regexp {^tcp:(.*)} $value {} s
	    set data($port,serv) $s
	}
    }
    unset ports(0)
    set portlist {}
    foreach port [array names ports] {
	if ![string match $service $ports($port)] continue
	set svc [InfoName $ports($port)]
	lappend portlist $port
	Get .services.$svc.description data($port,sdescription)
	Get .tcp.s$port.description data($port,description)
	foreach v $values {
	    if ![info exists data($port,$v)] {
		set data($port,$v) {}
	    }
	}
    }
    return [lsort -integer $portlist]
}

# Get information on one port, if it is configured
#
proc GetPortInfo {arrayvar port} {
    upvar $arrayvar data

    catch {unset data}
    set l {service protocol server start description timeout authmethod inetd}
    foreach v $l {
	if ![Get .tcp.s$port.service data($v)] {
	    set data($v) {}
	}
    }
    if ![string length $data(service)] {
	return 0
    }
    return 1
}

# Print the list of configured agents. Args are the same as for
# GetConfiguredPorts
#
proc ShowConfiguredPorts { {arrayvar {}} {service {}} } {
    if [string length $arrayvar] {
	upvar $arrayvar data
    }
    set ports [GetConfiguredPorts data $service]
    set fmt "%5s %5s %-8s %-18s %s"

    set n [llength $ports]
    if !$n {
	return 0
    }
    puts [format $fmt Port Proto Method Server Service]
    puts [line 79 =]
    foreach port $ports {
	set d $data($port,description)
	if ![string length $d] {
	    set d $data($port,sdescription)
	}
	if [string length $data($port,inetd)] {
	    set serv "(inetd $data($port,inetd))"
	} elseif ![string compare [string tolower $data($port,start)] hook] {
	    set serv "(bindhook)"
	} else {
	    set serv $data($port,serv)
	}
	puts [format $fmt \
		$port $data($port,protocol) \
		$data($port,authmethod) $serv \
		"$data($port,service) ($d)"]
    }
    return $n
}

# Remove a port configuration
#
proc DeletePort {port} {
    
    Get .tcp.s$port.inetd i
    if [string length $i] {
	catch {IWrap::Unwrap $i [CatDir APP]/bin/iwrap result}
    }
    

    

    DeleteNode .tcp.s$port
}

# Install a template file for a port
#
proc InstallPortTemplate {service port} {
    set dir [CatDir APP]
    set etc [CatDir ETC]

    set where $etc/tcp/s$port.cfg
    set node .tcp.s$port
    foreach f [list $dir/lib/$service/port.cfg $dir/lib/port.cfg] {
	if [file exists $f] {

	    
	    file copy -force $f $where
	    LoadFrom $node $where
	    

	     

	    return 1
	}
    }
    return 0
}

# Configure a port, with a bunch of arguments. No error checking.
# Overwrites any previous configuration.
#
proc ConfigurePort {port service protocol authmethod
		    server start description} {
    DeletePort $port
    InstallPortTemplate $service $port
    set p .tcp.s$port
    set server tcp:$server
    foreach v {service protocol authmethod server start description}  {
	Set $p.$v [set $v]
    }
    Set .tcp.s$port.inetd "" 1

    return 1
}

# Ditto when wrapping a service in inetd.conf
#
proc ConfigureInetdPort {tag service proto method start descr} {
    
    set port [GetServByName $tag]
    
    if !$port {
	return 0
    }
    set server "localhost:$port"
    set args [list $port $service $proto $method $server $start $descr]
    if ![eval ConfigurePort $args] {
	return 0
    }
    Set .tcp.s$port.inetd $tag
    if ![IWrap::Wrap $tag [CatDir APP]/bin/iwrap result] {
	return 0
    }
    return 1
    
}

# Ditto when "bindhooking" an application
#
proc ConfigureBindHookPort {port service proto method descr} {
    
    return 0
    
}




#----------------------------------------------------------------------
# Services
#

proc GetInstalledServices {arrayvar {name {}}} {
    upvar $arrayvar data

    if ![string length $name] {
	set name *
    } elseif ![regexp {.*-\d+\.\d+$} $name] {
	set name $name-*
    }
    catch {unset data}
    set services {}
    foreach node [List .services] {
	if ![regexp {.[^.]+.([^.]+).(.*)} $node {} service sub] continue
	if ![string match $name $service] continue
	Get $node value
	set data($service,$sub) $value
	lappend services $service
    }
    set tmp(0) 0
    unset tmp(0)
    foreach s $services {
	set tmp($s) 1
    }
    set services [array names tmp]
    foreach s $services {
	foreach n {description program} {
	    if ![info exists data($s,$n)] { set data($s,$n) {} }
	}
	if ![info exists data($s,name)] {
	    regsub -all {_} $s . data($s,name)
	}
    }
    return $services
}

#----------------------------------------------------------------------
# BoKS
#

proc GetBoksInfo {arrayvar} {
    upvar $arrayvar data

    set ok 0
    catch {unset data}
    set params {nodekey nodekeyfile domain servers cachettl myaddress logfile}
    foreach p $params {
	if [Get .global.boks.$p data($p)] {
	    set data($p,set) 1
	    incr ok
	} else {
	    set data($p,set) 0
	}
    }
    return $ok
}

proc ConfBoks {what value} {
    set b .global.boks
 
    if [string length $value] {
        Set $b.$what $value
    } else {
        Set $b.$what "" 1
    }
}

proc ConfBoksNodekey {{filename {}}} {
    if ![PasswordFile boks where] {
	puts stderr "Cannot write BoKS Node Key File"
	return 0
    }
    if [string length $filename] {
	if [catch {set f [open $filename]} ] {
	    Msg "Could not open password file: $filename"
	    return 0
	}
	set password [gets $f]
	close $f
    } else {
	if [catch {set password [exec catpsd -gQ "Confirm: " 2>@ stderr]}] {
	    return 0
	}
    }
    set f [open $where w]
    puts $f $password
    close $f
    ConfBoks nodekeyfile $where

    return 1
}

# See if we can talk to boks
#
proc TryBoks {{domain {}} {nodekey {}} {servers {}}} {
    if ![string length $domain] {
	Get .global.boks.domain domain
    }
    if ![string length $servers] {
	Get .global.boks.servers servers
    }
    if ![string length $nodekey] {
	if ![Get .global.boks.nodekey nodekey] {
	    if [Get .global.boks.nodekeyfile file] {
		if [catch {set f [open $file]}] {
		    return 0
		}
		gets $f nodekey
		close $f
	    }
	}
	if ![string length $nodekey] {
	    return 0
	}
    }
    set args [list -k$nodekey]
    if [string length $servers] {
	lappend args -s$servers
    }
    if [string length $domain] {
	lappend args -d$domain
    }
    if [catch {eval exec boks $args}] {
	return 0
    }
    return 1
}

#----------------------------------------------------------------------
# HostPSD
#
proc ConfHostPSD {filename { passwordfile {} }} {
    if ![string length $filename] {
	return 0
    }
    if ![file exists $filename] {
	puts stderr "$filename: does not exist"
	return 0
    }
    set filename [file nativename $filename]
    if ![PasswordFile psd where] {
	puts stderr "Cannot write the Virtual Card Password file"
	return 0
    }
    set arglist [list catpsd -tk$filename]
    if [string length $passwordfile] {
	set passwordfile [file nativename $passwordfile]
	if [catch {set f [open $passwordfile]} ] {
	    puts stderr "Could not open password file: $passwordfile"
	    return 0
	}
	set password [gets $f]
	close $f
	set f [open $where w]
	puts $f $password
	close $f
	lappend arglist -p$where
    } else {
	lappend arglist -o$where
    }
    lappend arglist 2>@ stderr

    if [catch {set type [eval exec $arglist]} msg] {
	D 1 "ConfHostPSD: $arglist : $msg"
	puts stderr "Failed to open Virtual Card file"
	return 0
    }
    set psd [CatDir ETC]/localhost.psd
    if [catch {file copy -force $filename $psd} msg] {
	D 1 "ConfHostPSD: $msg"
	puts stderr "Failed to install Agent Host Virtual Card file"
	return 0
    }
    set h .global.hostpsd
    Set $h.file [file nativename $psd]
    Set $h.password {} 1
    Set $h.passwordfile [file nativename $where]
    if {[string compare $type boks] == 0} {
	Set $h.boks 1
    } else {
	Set $h.boks 0
    }
    return 1
}

#----------------------------------------------------------------------
# ELS
#
proc ConfELS {servers} {
    set e .global.els

    if [string length $servers] {
	Set $e.servers $servers
    } else {
	Set $e.servers {} 1
    }
    return 1
}

proc GetELSInfo {arrayvar} {
    upvar $arrayvar data

    set e .global.els
    return [Get $e.servers data(servers)]
}

#----------------------------------------------------------------------
# List all global data
#
proc ShowCurrentConfig {{what {}}} {
    set fmt " %-22s: %s"
    set what [string tolower $what]
    if ![llength $what] {
	set what {dir els boks hostpsd ca}
    }
    if {[lsearch -glob $what dir*] > -1} {
	puts "\nDirectories\n[DashLine 70]"
	ShowData .global $fmt {
	    applicationroot "Application Programs"
	    etcroot	    "Configuration Files"
	    varroot	    "Data Files"
	}
    }
    if {[lsearch -glob $what els] > -1} {
	puts "\nELS\n[DashLine 70]"
	ShowData .global.els $fmt {
	    servers	"Server List"
	    logfile	"Log Spool File"
	}
    }
    if {[lsearch -glob $what boks] > -1} {
	puts "\nBoKS Setup\n[DashLine 70]"
	ShowData .global.boks $fmt {
	    domain 	"Domain Port Number"
	    nodekeyfile "Node Key File"
	    servers	"Server List"
	    myaddress	"Local Address"
	    logfile	"Log Spool File"
	}
    }
    if {[lsearch -glob $what host*] > -1} {
	puts "\nAgent Host Virtual Card\n[DashLine 70]"
	ShowData .global.hostpsd $fmt {
	    file	 "Virtual Card File"
	    passwordfile "Password File"
	}
    }
    if {[lsearch -glob $what ca] > -1} {
	puts "\nCA Storage\n[DashLine 70]"
	ShowData .global.ca $fmt {
	    dbfile	"CA Database"
	}
    }
}

proc ShowData {prefix fmt speclist} {
    foreach {n str} $speclist {
	Get $prefix.$n s
	puts [format $fmt $str $s]
    }
}

#----------------------------------------------------------------------
# CA admin functions
#

proc CaMenuImport {} {
    global Menus

    Hdr "Import CA certificate"
    PromptList importca Menus data {
	PromptFor file z "Enter filename of CA certificate:"
    } {
	Get .global.ca.dbfile dbfile
	set cafile $data(file,value)
	catch {exec cadb -v -d $dbfile -a $cafile 2>@ stderr}
    }
}

proc CaMenuDel {} {
    global Menus

    Hdr "Remove imported CA certificates"
    if ![ShowImportedCA] {
	PressReturn
	return 0
    }
    PromptList delca Menus data {
	PromptFor index z "Enter index of certificate to delete:"
    } {
	Get .global.ca.dbfile dbfile
	catch {set l [exec cadb -d $dbfile -l 2>@ stderr]}
	set l [split $l "\n"]
	set remidx $data(index,value)
	catch {set dn [lindex $l $remidx]}
	catch {exec cadb -v -d $dbfile -r $dn 2>@ stderr}
    }
}

proc CaMenuList {} {
    global Menus
    
    Hdr "List imported CA certificates"
    ShowImportedCA

    PressReturn
}

# Print the list of configured agents. Args are the same as for
# GetConfiguredPorts
#
proc ShowImportedCA {} {

    Get .global.ca.dbfile dbfile
    catch {set dnlist [exec cadb -v -d $dbfile -l 2>@ stderr]}

    if ![llength $dnlist] {
	puts "(No CA certificates imported)"
	return 0
    }
    set dnlist [split $dnlist "\n"]

    set fmt "%-7s %s"

    puts [format $fmt Index "Certificate Info"]
    puts [line 40 =]
    set i 0
    foreach dn $dnlist {
	puts [format $fmt $i $dn]
	incr i
    }
    return $i
}

} ;# end namespace CatConf
